----------------------------------------------------------------------------------------------------
-- DYNAMIC MUD SCRIPT
----------------------------------------------------------------------------------------------------
-- Purpose: To add mud to maps
-- Author:  reallogger
--
-- Copyright (c) Realismus Modding, 2019
----------------------------------------------------------------------------------------------------
-- add paintable mud by derelky
-- visibility added only when the ground is wet
--
--
----------------------------------------------------------------------------------------------------
DynamicMudvisual = {}
local DynamicMudvisual_mt = Class(DynamicMudvisual)
InitObjectClass(DynamicMudvisual, "DynamicMudvisual")

function DynamicMudvisual:new(isServer)
	local self = setmetatable({}, DynamicMudvisual_mt)
	self.isServer = isServer
	
		if self.isServer then
			g_currentMission.environment:addMinuteChangeListener(self)
		end
		local minute = 15
		DynamicMudvisual.minuteChanged(self, minute)
    return self
end

function DynamicMudvisual:delete()

	if self.isServer then
		g_currentMission.environment:removeMinuteChangeListener(self)
	end
	DynamicMudvisual:superClass().delete(self)
end

function DynamicMudvisual:minuteChanged(minute)
    if minute == nil then
	else	
		if minute % 15 == 0 then
			DynamicMudvisual:setmud()
		end
	end
end

function DynamicMudvisual:setmud(first)


			local nomud = false
			local statemud = getVisibility(TerrainControl.dirtyDensityId)
			if g_seasons ~= nil then
				if g_seasons.weather:isGroundFrozen() or g_currentMission.environment.weather:getGroundWetness() < 0.2 then
					nomud=true
				end
			else
				if g_currentMission.environment.weather:getGroundWetness() < 0.01 then
				nomud=true
				end
			end
			
			if not nomud == statemud then
			else
					DynamicMudvisual:changemud(nomud)
			    if self.isServer then
					g_server:broadcastEvent(senddynmudvis:new(nomud))
				end
			end

end

function DynamicMudvisual:changemud(nomud)
	local layerId = TerrainControl.dirtyDensityId
	if nomud then
		setVisibility(layerId, false)		
	else
		setVisibility(layerId, true)
	end
end

senddynmudvis = {}
local senddynmudvis_mt = Class(senddynmudvis, Event)

InitEventClass(senddynmudvis, "senddynmudvis")

function senddynmudvis:emptyNew()
    local self = Event:new(senddynmudvis_mt)

    return self
end

function senddynmudvis:new(nomud)
    local self = senddynmudvis:emptyNew()

    self.nomud = nomud

    return self
end

function senddynmudvis:writeStream(streamId, connection)
    streamWriteBool(streamId, self.nomud)
end

function senddynmudvis:readStream(streamId, connection)
    self.nomud = streamReadBool(streamId)

    self:run(connection)
end

function senddynmudvis:run(connection)
    if connection:getIsServer() then
        DynamicMudvisual:changemud(self.nomud)
    end
end







function load(mission)
	
    isServer = mission:getIsServer()
    dynamicMudvisual = DynamicMudvisual:new(isServer)	
	addModEventListener(dynamicMudvisual)


end

DynamicMud = {}

local directory = g_currentModDirectory
local modName = g_currentModName
local ModMap_mt = Class(ModMap, Mission00)

DynamicMud.FIRST_CHANNEL = 0
DynamicMud.NUM_CHANNELS = 1

DynamicMud.ALWAYS_MUD_CHANNEL = 1



function DynamicMud:loadMap(name)
end

function DynamicMud:loadSavegame()
end

function DynamicMud:saveSavegame()
end

function DynamicMud:update(dt)
end

function DynamicMud:mouseEvent(posX, posY, isDown, isUp, button)
end

function DynamicMud:keyEvent(unicode, sym, modifier, isDown)
end

function DynamicMud:draw()
end

function DynamicMud:delete()
end

function DynamicMud:deleteMap()
end

function DynamicMud.onTerrainLoaded()

	local densities = {	"dirty"};
	for i=1, #densities do
		local key = string.format("%sDensityId", densities[i]);
		if not TerrainControl[key] then
			TerrainControl[key] = Utils.getNoNil(getChild(g_currentMission.terrainRootNode, densities[i]), 0);
		end;
	end;

    DynamicMud.terrainSize = g_currentMission.terrainSize

    DynamicMud.modifierValue = DensityMapModifier:new(TerrainControl.dirtyDensityId, DynamicMud.FIRST_CHANNEL, DynamicMud.NUM_CHANNELS)
    DynamicMud.filterType = DensityMapFilter:new(TerrainControl.dirtyDensityId, DynamicMud.FIRST_CHANNEL, DynamicMud.NUM_CHANNELS)

end

function DynamicMud.installSpecializations()
    g_specializationManager:addSpecialization("TerrainControl", "TerrainControl", Utils.getFilename("maps/scripte/TerrainControl.lua", directory), nil)

    for typeName, typeEntry in pairs(g_vehicleTypeManager:getVehicleTypes()) do
        if SpecializationUtil.hasSpecialization(Wheels, typeEntry.specializations) then
            g_vehicleTypeManager:addSpecialization(typeName, modName .. ".TerrainControl")
        end
    end

end



function Vehicle:DynamicMud_getSpecTable(name)
    local spec = self["spec_" .. modName .. "." .. name]
    if spec ~= nil then
        return spec
    end

    return self["spec_" .. name]
end

addModEventListener(DynamicMud)

FSBaseMission.initTerrain = Utils.appendedFunction(FSBaseMission.initTerrain, DynamicMud.onTerrainLoaded)
Mission00.onStartMission = Utils.appendedFunction(Mission00.onStartMission, load)
VehicleTypeManager.validateVehicleTypes = Utils.prependedFunction(VehicleTypeManager.validateVehicleTypes, DynamicMud.installSpecializations)